#ifndef LOG_H
#define LOG_H

#include "mutex.h"
#include <sstream>
#include <fstream>
#include <ostream>
#include <vector>
#include <list>

// Class for outputing log and error messages to multiple destinations
class Log {
public:
    
    // defaults to true
    bool writeTimestamps;
    
    // Adds stream to the list of streams to output to
    Log(std::ostream &stream);
    
    // Adds filename to the list of files to output to
    Log(const std::string &filename);
    
    // Allow no-output Log instanciation
    Log();
    
    // Sets stream to recieve all messages sent to this log instance
    // until removeOutStream or the dtor is called.
    void addOutStream(std::ostream &stream);
    
    // Removes a stream previously added with addOutStream.  If the same
    // stream was added more then once the most recent addition
    // will be removed.
    void removeOutStream(std::ostream &stream);
    
    // Outputs all following messages to this file until removeOutFile
    // or the dtor is called.
    void addOutFile(const std::string &filename);
    
    // Stops outputing to filename.  If addOutFile was called for the
    // same file multiple times the most recent addition will be removed
    void removeOutFile(const std::string &filename);
    
    // Writes message to all output streams and files.  Only valid
    // datatypes may be used so "\n" is okay but std::endl is not.
    template<class T>
    void write(const T &message)
    {
        using namespace std;
        
        if(!writeTimestamps) {
            
            rawWrite(message);
        }
        else {
            
            ostringstream ss;
            string smessage;
            string timestamp;
            time_t secs;
            
            time(&secs);
            timestamp = "(";
            timestamp += ctime(&secs);
            timestamp.erase(timestamp.size() - 1); // ctime adds an extra
                                                   // newline that we don't
                                                   // want.
            timestamp += "): ";
            
            ss << message;
            
            smessage = ss.str();
            
            for(unsigned int i = 0; i < smessage.size(); i++) {
                
                if(!wroteTimestamp) {
                    
                    rawWrite(timestamp);
                    wroteTimestamp = true;
                }
                
                rawWrite(smessage.at(i));
                
                if(smessage.at(i) == '\n')
                    wroteTimestamp = false;
            }
        }
    }
    
    // Wrapper of write function allowing log << "My message"; usage
    template<class T>
    Log &operator << (const T &message)
    {
        write(message);
        
        return *this;
    }

private:
    
    template<class T>
    void rawWrite(const T &message)
    {
        using namespace std;
        
        for(unsigned int i = 0; i < outList.size(); i++) {
            
            *outList[i] << message;
        }
        
        for(list<FileInfo>::iterator itr = openFiles.begin(); itr != openFiles.end(); ++itr) {
            
            *itr->file << message;
        }
    }
    
    struct FileInfo {
        std::ofstream *file;
        std::string name;
        
        FileInfo(const std::string &filename)
            : file(new std::ofstream(filename.c_str())), name(filename)
        {
            
        }
        
        FileInfo(const FileInfo &f)
            : file(f.file), name(f.name)
        {
            const_cast<FileInfo&>(f).file = 0;
        }
        
        ~FileInfo()
        {
            delete file;
        }
    };
    
    std::list<FileInfo> openFiles;
    std::vector<std::ostream*> outList;
    
    bool wroteTimestamp;
};

// Two default log wrappers for cerr and cout.
extern Log err;
extern Log out;

#endif
