#ifndef PRINT_H
#define PRINT_H

#include "bitmap.h"
#include "font.h"
#include <sstream>
#include <string>

// See the Print class below's comments for a description of how to use it

// Used internaly by Print
namespace PrintOpHelp {
    
    static const int printAction = 1;
    static const int colorAction = 2;
    
    template<class T>
    struct OperatorParser {
        
        // Prints message using printer
        static int parseAction(std::string &output, Tint&, const T &message)
        {
            std::ostringstream ss;
            
            ss << message;
            
            output = ss.str();
            
            return printAction;
        }
    };
    
    template< >
    struct OperatorParser <Tint> {
        
        // Changes the color state of printer to tint
        static int parseAction(std::string&, Tint &printerColor, const Tint &newColor)
        {
            printerColor = newColor;
            
            return colorAction;
        }
    };
}

// Prints output to the screen.
// Example usage:
//
// Print(10, 10) << "Frames per second: " << fps;
// Print(10, 100) << "Some multiline text:\nLine Two\nLine Three.";
// Print(10, 200, Tint(1, 0, 0)) << "Some red text.";
// Print(10, 300, Tint(1, 1, 1, .5)) << "Half transparent text.";
// Print(10, 400) << Tint(1, 0, 0) << "Red text" << Tint() << ", "
//                << Tint(0, 1, 0) << "green text" << Tint() << ", "
//                << Tint(0, 0, 1) << "blue text" << Tint() << " and "
//                << Tint(1, 1, 1, .5) << "half transparent text.";
class Print {
public:
    
    enum Align { LEFT, CENTER, RIGHT } alignment;
    
    // Specifies the upperlefthand side of the text output and the outputting
    // color (tint).  The color defaults to white with full opacity.
    Print(int x, int y, const Tint &tint = Tint(), Align alignment = LEFT);
    
    // Prints message to the screen or sets the printer color to the tint
    // supplied in message.
    template<class T>
    Print &operator <<(const T &message)
    {
        using namespace PrintOpHelp;
        
        switch(OperatorParser<T>::parseAction(nextMessage, tint, message))
        {
            case printAction:
                print(nextMessage);
                break;
        }
        
        return *this;
    }
    
    // Prints the string to the screen using the appropriate driver.
    // interprets newlines by incrementing y by one and one third of the
    // text height.
    // Also increments x by the length of the text.
    //
    // Operator << provides a nice interface to this method
    void print(std::string message);
    
    // The current tint of the text.
    //
    // Operator << provides a nice interface to modifying this attribute
    Tint tint;
    
    // Returns a reference to the font renderer being used internaly.
    static FontRenderer &getFontRenderer();
    
    // init must be called before using any instance of Print and shutdown
    // must be called once you're done using the Print class
    static void init();
    static void shutdown();
    
    // Non-blocking version of init.  After calling this method call pollInit untill it returns false
    static void startInit();
    
    // Call startInit before using this method.  Keep calling it untill it returns false.
    static bool pollInit();
    
private:
    
    int x, y;
    int startingX;
    
    // Used by operator << to store the next message to be outputed
    std::string nextMessage;
};

#endif
