#ifndef SOUNDS_H
#define SOUNDS_H

#include "datafile.h"

// Routines for accessing sounds in a datafile.  The sounds should be
// structured in the datafile so that objdir can be postfixed with a
// number starting from 0 and going up until there is no object with
// that name found.  The sounds should also be given english style
// names under the SOND property of the datafile object.
// Example datafile structure:

/*
 * Name    | Properties
 * sfx/0     SOND=gunshot
 * sfx/1     SOND=music
 * sfx/2     SOND=uiAccept
 */

// Example usage:
//
// Sounds::load();
//
// ...
//
// Sounds::loop("music");
//
// if(fire)
//   Sounds::play("gunshot", PlayInfo(player_x / 640.0));
//
// if(accept)
//   Sounds::play("uiAccept");
//
// if(die)
//   Sounds::stop("music");
//
// if(quit)
//   Sounds::stopAll();
namespace Sounds {
    
    // A value of 1 is full volume and a value of 0 is silent
    void setSfxVolume(float volume);
    
    // A value of 1 is full volume and a value of 0 is silent
    float getSfxVolume();
    
    // Information describing how to play the sound
    struct PlayInfo {
        float pan; // 0 being left, .5 being center, 1 being right
        float frequency; // 1 being regular, 2 being double the frequency
        
        // Specify default values for all values
        PlayInfo(float pan_ = .5, float vol = 1, float freq = 1)
            : pan(pan_), frequency(freq) {}
    };
    
    // Loads the sounds out of the datafile and into memory.  If the
    // datafile is already loaded then stores pointers to the samples.
    // objdir specifies the directory in the datafile where the sounds
    // may be found.
    void load(const std::string &objdir = "sfx/", datafile_t *datailfe = &datafile);
    
    // Call this before shutting down allegro
    void unload();
    
    // Finds the song named 'soundName' (soundName is the property in
    // the datafile object named 'SOND') and plays it.  Returns true
    // if the sample was found and played or false if it could not be
    // played for some reason.  See the comments for PlayInfo for a
    // description of how this is used.
    bool play(const std::string &soundName, const PlayInfo &playingInfo = PlayInfo());
    
    // Finds the song named 'soundName' (soundName is the property in
    // the datafile object named 'SOND') and plays it in a loop.  Returns
    // true if the sample was found and played or false if it could not be
    // played for some reason.  See the comments for PlayInfo for a
    // description of how this is used.
    bool loop(const std::string &soundName, const PlayInfo &playingInfo = PlayInfo());
    
    // Used to adjust the PlayInfo of a sound while its playing.  Returns
    // true if the sample was found or false if the sound could not be
    // adjusted for some reason.  See the comments for PlayInfo for a
    // description of how this is used.
    //
    // Usefull for manipulating looped sounds while they're playing.  The
    // looped parameter will change the looped value of the sound.  Setting
    // this to 0 will make a looped sound stop at the end of its next loop.
    bool adjust(const std::string &soundName, bool looped = 0, const PlayInfo &playInfo = PlayInfo());
    
    // Stops all currently playing instances of soundName.
    void stop(const std::string &soundName);
    
    // Stops all instances of all sounds.
    void stopAll();
}

#endif
