#ifndef WIDGETS_LISTBOX_H
#define WIDGETS_LISTBOX_H

#include "widget.h"
#include "animation.h"
#include "log.h"
#include <vector>

namespace widgets {
    
    // This listbox is not meant to usable directly from GUI scripts, insted
    // it is meant to be inherited from
    class ListBox : public Widget {
    public:
        
        // Pointer to the active ListBox, set to 'this' before calling Element's
        // on* event methods.
        static ListBox *active;
        
        ListBox();
        virtual ~ListBox() { }
        
        // This class ignores all attributes
        
        class Element {
        public:
            
            virtual ~Element() { }
            
            // The height of this element
            int h;
            
            // Draws the element to the screen, sx and sy specify where on the
            // source 'element' the drawing should start.  0 and 0 would mean
            // a complete draw and 1 and 1 would mean start one pixel down
            // and right from the top-left corner.
            //
            // dx and dy specify where on the screen the element should be drawn
            // w and h specify how much of the element should be drawn.  The
            // remaining portions of the element that go past w and h should be
            // clipped off.
            // 
            // tint specifies any tinting that should be applied to the drawing
            virtual void draw(int sx, int sy, int dx, int dy, int w, int h, const Tint &tint = Tint()) { }
            
            // Actions this element needs to recieve.  x and y specify the mouse
            // coords on the element with 0, 0 being the upper-left corner of
            // the element's location.
            virtual Action onHover(int x, int y) { return OK; }
            virtual Action onLeave() { return OK; }
            virtual Action onClick(int x, int y) { return OK; }
            virtual Action onRelease(int x, int y) { return OK; }
            virtual Action onIdle() { return OK; }
        };
        
        std::vector<Element*> elements;
        
        // Returns the element that is lower then 'element' or 'element' if it is already
        // the lowest one.  If 'element' cannot be found in elements 'element' is returned
        Element *getLower(Element *element);
        
        // Returns the element that is higher then 'element' or 'element' if it is already
        // the highest one.  If 'element' cannot be found in elements 'element' is returned
        Element *getHigher(Element *element);
        
        // Modifies scroll such that 'element' is viewable if it was previously hidden by the
        // scroll offset.  If 'element' is not in the list of elements, scrollTo does nothing.
        void scrollTo(Element *element);
        
    protected:
        
        // Called when the widget needs to be drawn.  xstart and ystart are added to the widget's
        // x / y position to deduce the drawing position.
        virtual void draw(int xstart, int ystart, const Tint &tint);
        
        // The events this widget needs to recieve
        virtual Action onHover();
        virtual Action onLeave();
        virtual Action onClick();
        virtual Action onRelease();
        virtual Action onIdle();
        
    private:
        
        Element *hasHover;
        
        int scroll;
        
        int xstart, ystart;
        
        int mouseWheelPos;
        
        bool hasMouse;
        
        animation scrollbar;
        animation scrollpoint;
        
        animation buttonActive;
        animation buttonInactive;
        animation buttonHover;
        
        animation *up, *down;
        
        // Returns true if the mouse is over the up button
        bool overUp();
        
        // Returns true if the mouse is over the down button
        bool overDown();
        
        // Returns true if the mouse is over the scroll bar
        bool overScrollbar();
        
        // Sets scroll according to the position of the mouse.  Does no
        // checking to see if mouse is actually in place to handle
        // scrolling scroll bar.  That should be checked before calling
        // this funciton.
        void setScroll();
        
        // Checks and puts scroll into proper range of values.
        void wrapScroll();
        
        // Returns the height in pixels of all the internal elements plus padding combined
        int getInsideHeight();
    };
}

#endif
