#include "socket.h"
#include "log.h"
#include "servhandler.h"
#include "thread.h"
#include "mutex.h"
#include "world.h"
#include <queue>
#include <vector>
#include <map>
#include <signal.h>
using namespace std;
using namespace net;

map<Socket::Sock, unsigned int> ids;
vector<Socket> clients;

namespace con {
    
    // This mutex must be locked to use this namespace
    Mutex mutex;
    
    struct Client {
        Socket socket;
        string ip;
        unsigned short port;
    };
    
    // List of recently connected clients
    queue<Client> newClients;
}

int getConnections(void*)
{
    Socket server(Socket::TCP);
    
    if(!server.bind(1423))
        return 1;
    
    if(!server.listen())
        return 1;
    
    bool status;
    
    for(;;) {
        
        string ip;
        unsigned short port;
        
        Socket sock = server.accept(&status, &ip, &port);
        
        if(!status)
            return getErrno();
        
        con::Client client = { sock, ip, port };
        
        con::mutex.lock(); {
            
            con::newClients.push(client);
        
        } con::mutex.unlock();
    }
    
    return 0;
}

bool shouldKillit = false;

void killit(int)
{
    shouldKillit = true;
}

int main(int argc, char *argv[])
{
    out.addOutFile("serverlog.txt");
    err.addOutFile("servererrors.txt");
    
    bool loaded = 0;
    
    try {
        out << "WWo server starting.\n";
        
        world.itemDb.load("world.idb");
        world.resourceDb.load("world.rdb");
        world.buildingDb.load("world.bdb");
        world.trainDb.load("world.tdb");
        world.loadCities("world.cit");
        world.loadResources("world.res");
        world.loadTracks("world.trk");
        world.loadTrains("world.trn");
        world.loadBuildings("world.bld");
        world.loadTerminals("world.trl");
        world.loadUsers("world.usr");
        loaded = 1;
        
        net::init();
        
        signal(SIGINT, killit); // Control-C
#ifndef WIN32
        signal(SIGTSTP, killit); // Control-Z
#endif
        signal(SIGABRT, killit); // bad stuff
        
        Thread t(getConnections);
        
        while(!shouldKillit) {
            
            if(t.done()) {
                
                err << "Server error: " << getError(t.returnCode()) << "\n";
                break;
            }
            
            con::mutex.lock(); {
                
                if(!con::newClients.empty()) {
                    
                    con::Client client = con::newClients.back();
                    con::newClients.pop();
                    
                    clients.push_back(client.socket);
                    
                    out << "Client " << clients.size() - 1 << " connected from " << client.ip << ":"
                        << client.port << "\n";
                }
            
            } con::mutex.unlock();
            
            for(unsigned int i = 0; i < clients.size(); i++) {
                
                tagHandlers[0](0, Socket(clients[7], false));
            }
            
            bool ready = false;
            Socket socket;
            
            if(!clients.empty() && !net::select(&clients[0], &clients[0] + clients.size(), &ready, &socket))
                err << "Server select error: " << getError() << "\n";
            else if(ready) {
                
                char buf[1024];
                char tag;
                int size = 0;
                int recieved = 0;
                int i = 0;
                
                i = socket.recv(&tag, 1);
                
                if(i < 1)
                    err << "Server recv error: " << getError() << "\n";
                else {
                    
                    if(tag > maxTag || tag < 1) {
                        
                        i = -1;
                        err << "Invalid tag " << (int)tag << "\n";
                    }
                    else {
                        size = sizeOfData[(int)tag];
                        
                        while(recieved < size) {
                            
                            i = socket.recv(buf + recieved, size - recieved);
                            
                            if(i > 0)
                                recieved += i;
                            else {
                                
                                err << "Server recv error: " << getError() << "\n";
                                break;
                            }
                        }
                    }
                }
                
                if(socket.closed || i < 0) {
                    
                    vector<Socket>::iterator itr;
                    
                    for(itr = clients.begin(); itr != clients.end(); ++itr)
                        if(itr->sock == socket.sock) {
                            
                            ids.erase(itr->sock);
                            
                            out << "Client " << &*itr - &clients[0] << " disconnected.\n";
                            clients.erase(itr);
                            break;
                        }
                }
                else if(i > 0) {
                    
                    out << "Tag: " << (int)tag << "\n";
                    
                    tagHandlers[(int)tag](buf, socket);
                }
            }
        }
        
        net::shutdown();
    }
    catch (const char *error)
    {
        err << "Exception raised: " << error << "\n";
    }
    
    if(loaded) {
        
        world.saveCities("world.cit");
        world.saveResources("world.res");
        world.saveTracks("world.trk");
        world.saveTrains("world.trn");
        world.saveBuildings("world.bld");
        world.saveTerminals("world.trl");
        world.saveUsers("world.usr");
    }
    
    out << "WWo sever shutting down.\n";
}
