#include "bitmap.h"
#include "log.h"
#include "conf.h"
#include "software.h"
#include <list>
#include <cmath>
#include <map>

#ifndef NO_PTHREADS
#include <pthread.h>
#endif
using namespace std;

namespace {
    
#ifndef NO_PTHREADS
    pthread_mutex_t loadingLock;
    pthread_mutexattr_t attr;
#endif
    
    list<Bitmap*> instances;
}

Bitmap::Bitmap()
    : copiedFrom(0), w(0), h(0), bitmap(0), isInVram(0), texture(0), src(0)
{
    instances.push_back(this);
}

Bitmap::Bitmap(BITMAP *bsrc)
    : copiedFrom(0), w(0), h(0), bitmap(0), isInVram(0), texture(0), src(bsrc)
{
    if(!src)
        throw "Given invalid Allegro bitmap to Bitmap::Bitmap().";
    
    lockLoading();
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        loadBitmapHardware(src);
    else
        loadBitmapSoftware(src);
    
    unlockLoading();
    
    instances.push_back(this);
}

Bitmap::Bitmap(const Bitmap &csrc)
    : copiedFrom(0), w(csrc.w), h(csrc.h), bitmap(csrc.bitmap), isInVram(csrc.isInVram), texture(csrc.texture),
      src(csrc.src)
{
    csrc.copiedFrom = true;
    
    instances.push_back(this);
}

Bitmap &Bitmap::operator =(const Bitmap &src)
{
    copiedFrom = 0;
    w = src.w;
    h = src.h;
    bitmap = src.bitmap;
    isInVram = src.isInVram;
    texture = src.texture;
    this->src = src.src;
    
    src.copiedFrom = true;
    
    instances.push_back(this);
    
    return *this;
}

Bitmap::~Bitmap()
{
    if(!copiedFrom) {
        
        if(texture)
            ;//unloadBitmapHardware();
        
        if(bitmap)
            ;//unloadBitmapSoftware();
    }
    
    instances.remove(this);
}

void Bitmap::init()
{
#ifndef NO_PTHREADS
    pthread_mutexattr_init(&attr);
    pthread_mutex_init(&loadingLock, &attr);
#endif
}

void Bitmap::shutdown()
{
#ifndef NO_PTHREADS
    pthread_mutex_destroy(&loadingLock);
#endif
}

void Bitmap::unloadAll()
{
    for(list<Bitmap*>::iterator itr = instances.begin(); itr != instances.end(); ++itr) {
        
        if((*itr)->copiedFrom)
            continue;
        
        if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
            (*itr)->unloadBitmapHardware();
        else
            (*itr)->unloadBitmapSoftware();
    }
}

void Bitmap::reloadAll()
{
    for(list<Bitmap*>::iterator itr = instances.begin(); itr != instances.end(); ++itr) {
        
        if((*itr)->copiedFrom)
            continue;
        
        if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
            (*itr)->loadBitmapHardware((*itr)->src);
        else
            (*itr)->loadBitmapSoftware((*itr)->src);
    }
}

void Bitmap::lockLoading()
{
#ifndef NO_PTHREADS
    pthread_mutex_lock(&loadingLock);
#endif
}

void Bitmap::unlockLoading()
{
#ifndef NO_PTHREADS
    glFlush();
    
    pthread_mutex_unlock(&loadingLock);
#endif
}

namespace {
    
    int xoffset = 0;
    int yoffset = 0;
}

void Bitmap::gradient(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        gradientHardware(x, y, w, h, upperLeft, upperRight, lowerRight, lowerLeft);
    else
        gradientSoftware(x, y, w, h, upperLeft, upperRight, lowerRight, lowerLeft);
}

void Bitmap::line(int x1, int y1, int x2, int y2, int width, const Tint &left, const Tint &right)
{
    x1 += xoffset;
    x2 += xoffset;
    y1 += yoffset;
    y2 += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        lineHardware(x1, y1, x2, y2, width, left, right);
}

void Bitmap::scissorStart(int x, int y, int w, int h)
{
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        scissorStartHardware(x, y, w, h);
}

void Bitmap::scissorEnd()
{
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        scissorEndHardware();
}

void Bitmap::setShake(int xoff, int yoff)
{
    xoffset = xoff;
    yoffset = yoff;
}

void Bitmap::getShake(int *xoff, int *yoff)
{
    if(xoff)
        *xoff = xoffset;
    
    if(yoff)
        *yoff = yoffset;
}

void Bitmap::screenshot(const string &filename)
{
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        screenshotHardware(filename);
}

void Bitmap::blitToScreen(int x, int y, const Tint &tint, bool disableBlending)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        blitToScreenHardware(x, y, tint, disableBlending);
    else
        blitToScreenSoftware(x, y, tint, disableBlending);
    
    x -= xoffset;
    y -= yoffset;
}

void Bitmap::blitToScreend(double x, double y, const Tint &tint, bool disableBlending)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        blitToScreenHardwared(x, y, tint, disableBlending);
    else
        blitToScreenSoftwared(x, y, tint, disableBlending);
}

void Bitmap::circleBlitToScreen(int x, int y, int radius, float angle1, float angle2, const Tint &tint, bool disableBlending)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        circleBlitToScreenHardware(x, y, radius, angle1, angle2, tint, disableBlending);
}

void Bitmap::tileBlitToScreen(int x, int y, float w, float h, float xoffset, float yoffset, const Tint &tint, bool disableBlending)
{
    x += ::xoffset;
    y += ::yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        tileBlitToScreenHardware(x, y, w, h, xoffset, yoffset, tint, disableBlending);
    else
        tileBlitToScreenSoftware(x, y, w, h, xoffset, yoffset, tint, disableBlending);
}

void Bitmap::specialBlitToScreen(int x, int y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        specialBlitToScreenHardware(x, y, wchange, hchange, rotation, tint, disableBlending);
    else
        specialBlitToScreenSoftware(x, y, wchange, hchange, rotation, tint, disableBlending);
}

void Bitmap::specialBlitToScreend(double x, double y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending)
{
    x += xoffset;
    y += yoffset;
    
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        specialBlitToScreenHardwared(x, y, wchange, hchange, rotation, tint, disableBlending);
    else
        specialBlitToScreenSoftware((int)x, (int)y, wchange, hchange, rotation, tint, disableBlending);
}

GLuint Bitmap::getTexture()
{
    if(globalConfInstance->getRenderingMode() == Conf::HARDWARE)
        return texture;
    
    return 0;
}

BITMAP *Bitmap::getBitmap()
{
    if(globalConfInstance->getRenderingMode() == Conf::SOFTWARE)
        return bitmap;
    
    return NULL;
}

void Bitmap::gradientHardware(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft)
{
    glDisable(GL_TEXTURE_2D);
    
    glBegin(GL_QUADS);
        glColor4f(upperLeft.r, upperLeft.g, upperLeft.b, upperLeft.a);
        glTexCoord2i(0, 1);
        glVertex2i(x, y);
        
        glColor4f(upperRight.r, upperRight.g, upperRight.b, upperRight.a);
        glTexCoord2i(1, 1);
        glVertex2i(x + w, y);
        
        glColor4f(lowerRight.r, lowerRight.g, lowerRight.b, lowerRight.a);
        glTexCoord2i(1, 0);
        glVertex2i(x + w, y + h);
        
        glColor4f(lowerLeft.r, lowerLeft.g, lowerLeft.b, lowerLeft.a);
        glTexCoord2i(0, 0);
        glVertex2i(x, y + h);
    glEnd();
    
    glEnable(GL_TEXTURE_2D);
}

void Bitmap::lineHardware(int x1, int y1, int x2, int y2, int width, const Tint &left, const Tint &right)
{
    glDisable(GL_TEXTURE_2D);
    
    glLineWidth(width);
    
    glBegin(GL_LINES);
        glColor4f(left.r, left.g, left.b, left.a);
        glTexCoord1i(0);
        glVertex2i(x1, y1);
        
        glColor4f(right.r, right.g, right.b, right.a);
        glTexCoord1i(1);
        glVertex2i(x2, y2);
    glEnd();
    
    glEnable(GL_TEXTURE_2D);
}

void Bitmap::scissorStartHardware(int x, int y, int w, int h)
{
    glEnable(GL_SCISSOR_TEST);
    
    glScissor(x * SCREEN_W / 640, (480 - y - h) * SCREEN_H / 480, w * SCREEN_W / 640, h * SCREEN_H / 480);
}

void Bitmap::scissorEndHardware()
{
    glDisable(GL_SCISSOR_TEST);
}

void Bitmap::screenshotHardware(const string &filename)
{
    typedef unsigned char byte;
    
	BITMAP *bmp = create_bitmap(SCREEN_W,SCREEN_H);
    
	if(!bmp)
        throw "Out of memory.";
    
	byte *buff = new byte[4 * SCREEN_W * SCREEN_H];
    
	if(!buff) {
        
		destroy_bitmap(bmp);
		return;
	}

	glReadPixels(0, 0, SCREEN_W, SCREEN_H, GL_RGBA, GL_UNSIGNED_BYTE, buff);

	for(int y = 0; y < SCREEN_H; y++)
		memcpy(bmp->line[y],buff + 4 * SCREEN_W * (SCREEN_H - 1 - y), 4 * SCREEN_W);

	save_bitmap(filename.c_str(), bmp, NULL);

	destroy_bitmap(bmp);
	delete buff;
}

namespace {
    
    struct BitmapInfo {
        BITMAP *bitmap;
        bool isInVram;
        GLuint texture;
        int w, h;
        int instanceCount;
    };
    
    map<BITMAP*, BitmapInfo> textureLookup;
}

void Bitmap::loadBitmapHardware(BITMAP *src)
{
    map<BITMAP*, BitmapInfo>::iterator loc = textureLookup.find(src);
    
    if(loc != textureLookup.end()) {
        
        texture = loc->second.texture;
        w = loc->second.w;
        h = loc->second.h;
        
        loc->second.instanceCount++;
    }
    else {
        
        texture = allegro_gl_make_texture(src);
        
        w = src->w;
        h = src->h;
        
        if(!texture)
            throw "Unable to make texture from Allegro bitmap.";
        
        map<BITMAP*, BitmapInfo>::iterator loc = textureLookup.insert(make_pair(src, BitmapInfo())).first;
        
        loc->second.texture = texture;
        loc->second.w = w;
        loc->second.h = h;
        loc->second.instanceCount = 1;
    }
}

void Bitmap::unloadBitmapHardware()
{
    map<BITMAP*, BitmapInfo>::iterator itr = textureLookup.begin();
    
    if(itr == textureLookup.end())
        return;
    
    while(itr->second.texture != texture) {
        
        ++itr;
        
        if(itr == textureLookup.end())
            return;
    }
    
    if(!--(itr->second.instanceCount)) {
        
        glDeleteTextures(1, &itr->second.texture);
        textureLookup.erase(itr);
    }
    
    out << "Textures left: " << textureLookup.size() << "\n";
}

void Bitmap::tileBlitToScreenHardware(int x, int y, float w, float h, float xoffset, float yoffset, const Tint &tint, bool disableBlending)
{
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    
    glBegin(GL_QUADS);
        glTexCoord2f(xoffset, yoffset);
        glVertex2f(x, y + h);
        
        glTexCoord2f(xoffset, h / this->h + yoffset);
        glVertex2i(x, y);
        
        glTexCoord2f(w / this->w + xoffset, h / this->h + yoffset);
        glVertex2f(x + w, y);
        
        glTexCoord2f(w / this->w + xoffset, yoffset);
        glVertex2f(x + w, y + h);
    glEnd();
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);		
}

void Bitmap::blitToScreenHardware(int x, int y, const Tint &tint, bool disableBlending)
{
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    glBegin(GL_QUADS);
        glTexCoord2i(0, 1);
        glVertex2i(x, y);
        glTexCoord2i(1, 1);
        glVertex2i(x + w, y);
        glTexCoord2i(1, 0);
        glVertex2i(x + w, y + h);
        glTexCoord2i(0, 0);
        glVertex2i(x, y + h);
    glEnd();
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

}

void Bitmap::blitToScreenHardwared(double x, double y, const Tint &tint, bool disableBlending)
{
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    glBegin(GL_QUADS);
        glTexCoord2i(0, 1);
        glVertex2d(x, y);
        glTexCoord2i(1, 1);
        glVertex2d(x + w, y);
        glTexCoord2i(1, 0);
        glVertex2d(x + w, y + h);
        glTexCoord2i(0, 0);
        glVertex2d(x, y + h);
    glEnd();
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void Bitmap::specialBlitToScreenHardware(int x, int y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending)
{
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    float w = this->w * wchange;
    float h = this->h * hchange;
    
    glPushMatrix();
    
    glTranslatef(x + w / 2, y + h / 2, 0);
    glRotatef(rotation * 360 + 180, 0, 0, 1);
    
    glBegin(GL_QUADS);
        glTexCoord2i(1, 0);
        glVertex2f(-w/2, -h/2);
        glTexCoord2i(1, 1);
        glVertex2f(-w/2, h/2);
        glTexCoord2i(0, 1);
        glVertex2f(w/2, h/2);
        glTexCoord2i(0, 0);
        glVertex2f(w/2, -h/2);
    glEnd();
    
    glPopMatrix();
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);		
}

void Bitmap::specialBlitToScreenHardwared(double x, double y, float wchange, float hchange, float rotation, const Tint &tint, bool disableBlending)
{
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    float w = this->w * wchange;
    float h = this->h * hchange;
    
    glPushMatrix();
    
    glTranslated(x + w / 2, y + h / 2, 0);
    glRotatef(rotation * 360 + 180, 0, 0, 1);
    
    glBegin(GL_QUADS);
        glTexCoord2i(1, 0);
        glVertex2f(-w/2, -h/2);
        glTexCoord2i(1, 1);
        glVertex2f(-w/2, h/2);
        glTexCoord2i(0, 1);
        glVertex2f(w/2, h/2);
        glTexCoord2i(0, 0);
        glVertex2f(w/2, -h/2);
    glEnd();
    
    glPopMatrix();
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void Bitmap::circleBlitToScreenHardware(int x, int y, int radius, float angle1, float angle2, const Tint &tint, bool disableBlending)
{
    float sinR, cosR;
    
    if(angle1 >= angle2 || isinf(angle1) || isinf(angle2) || isnan(angle1) || isnan(angle2))
        return;
    
    glColor4f(tint.r, tint.g, tint.b, tint.a);
    
    if(disableBlending)
        glDisable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    glBindTexture(GL_TEXTURE_2D, texture);
    
    glBegin(GL_TRIANGLE_FAN);
        glTexCoord2f(0.5f, 0.5f);
        glVertex2f(x, y);
        
        for(float i = angle2; i > angle1; i -= M_PI / 180) {
            
            sinR = sinf(i);
            cosR = cosf(i);
            
            glTexCoord2f(sinR * 0.5f + 0.5f, cosR * 0.5f + 0.5f);
            glVertex2f(x + sinR * radius, y + cosR * radius);
        }
    glEnd();
    
    if(disableBlending)
        glEnable(GL_BLEND);
    
    if(tint.additive)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void Bitmap::gradientSoftware(int x, int y, int w, int h, const Tint &upperLeft, const Tint &upperRight, const Tint &lowerRight, const Tint &lowerLeft)
{
    Tint tint;
    
    // We cant actually do a gradient so just average the four colors into one
    tint.r = (upperLeft.r + upperRight.r + lowerRight.r + lowerLeft.r) / 4;
    tint.g = (upperLeft.g + upperRight.g + lowerRight.g + lowerLeft.g) / 4;
    tint.b = (upperLeft.b + upperRight.b + lowerRight.b + lowerLeft.b) / 4;
    tint.a = (upperLeft.a + upperRight.a + lowerRight.a + lowerLeft.a) / 4;
    
    set_trans_blender(0xff, 0xff, 0xff, (int)(tint.a * 0xff));
    drawing_mode(DRAW_MODE_TRANS, 0, 0, 0);
    rectfill(softwareRenderer.screen, x, y, x + w, y + h, makeacol32((int)(tint.r * 0xff), (int)(tint.g * 0xff), (int)(tint.b * 0xff), (int)(tint.a * 0xff)));
    drawing_mode(DRAW_MODE_SOLID, 0, 0, 0);
}

namespace {
    
    map<BITMAP*, BitmapInfo> bitmapLookup;
}

void Bitmap::loadBitmapSoftware(BITMAP *src)
{
    bitmap = 0;
    w = src->w;
    h = src->h;
    
    map<BITMAP*, BitmapInfo>::iterator loc = bitmapLookup.find(src);
    
    if(loc != bitmapLookup.end()) {
        
        bitmap = loc->second.bitmap;
        w = loc->second.w;
        h = loc->second.h;
        isInVram = loc->second.isInVram;
        
        loc->second.instanceCount++;
    }
    else {
        
        map<BITMAP*, BitmapInfo>::iterator loc = bitmapLookup.insert(make_pair(src, BitmapInfo())).first;
        
        loc->second.isInVram = isInVram = false;
        loc->second.w = w;
        loc->second.h = h;
        loc->second.instanceCount = 1;
        
        bitmap = src;
        
        loc->second.bitmap = bitmap;
    }
}

void Bitmap::unloadBitmapSoftware()
{
    map<BITMAP*, BitmapInfo>::iterator itr = bitmapLookup.begin();
    
    if(itr == bitmapLookup.end())
        return;
    
    while(itr->second.bitmap != bitmap) {
        
        ++itr;
        
        if(itr == bitmapLookup.begin())
            return;
    }
    
    if(!--(itr->second.instanceCount)) {
        
        if(itr->second.isInVram)
            destroy_bitmap(bitmap);
        
        bitmapLookup.erase(itr);
    }
}

void Bitmap::blitToScreenSoftware(int x, int y, const Tint &tint, bool disableBlending)
{
    if(disableBlending)
        draw_sprite(softwareRenderer.screen, bitmap, x, y);
    else if(tint.a < 1) {
        set_trans_blender((int)(tint.r * 0xff), (int)(tint.g * 0xff), (int)(tint.b * 0xff), (int)(tint.a * 0xff));
        draw_trans_sprite(softwareRenderer.screen, bitmap, x, y);
    }
    else if(tint.a > 0) {
        
        set_trans_blender((int)(tint.r * 0xff), (int)(tint.g * 0xff), (int)(tint.b * 0xff), (int)(tint.a * 0xff));
        draw_lit_sprite(softwareRenderer.screen, bitmap, x, y, 0xff - (int)(MIN(MIN(tint.r, tint.g), tint.b) * 0xff));
    }
}

void Bitmap::blitToScreenSoftwared(double x, double y, const Tint &tint, bool disableBlending)
{
    blitToScreenSoftware((int)x, (int)y, tint, disableBlending);
}

void Bitmap::tileBlitToScreenSoftware(int x, int y, float w_, float h_, float xoff, float yoff, const Tint&, bool)
{
    xoff -= (int)xoff;
    yoff -= (int)yoff;
    
    if(xoff < 0)
        xoff += 1;
    if(yoff < 0)
        yoff += 1;
    
    int s_x = (int)((1 - xoff) * geth());
    int s_y = (int)((1 - yoff) * getw());
    int d_x = x;
    int d_y = y;
    int w = (int)w_ + 1;
    int h = (int)h_ + 1;
    
    int tile_width = getw();
    int tile_height = geth();
    
    for(int tx = s_x / tile_width; tx < w && tx * tile_width - s_x < w; tx++) {
        for(int ty = s_y / tile_height; ty < h && ty * tile_height - s_y < h; ty++) {
            
            int des_x = ((tile_width * tx)  - s_x) + d_x;
            int des_y = ((tile_height * ty) - s_y) + d_y;
            
            int x_add = MAX(0, d_x - des_x);
            int y_add = MAX(0, d_y - des_y);
            
            masked_blit(bitmap, softwareRenderer.screen, x_add, y_add, des_x + x_add, des_y + y_add, tile_width + MIN(0, (w + d_x) -
                (des_x + tile_width)), tile_height + MIN(0, (h + d_y) - (des_y + tile_height)));
        }
    }
}

void Bitmap::specialBlitToScreenSoftware(int x, int y, float wchange, float hchange, float rotation, const Tint&, bool)
{
    // Allegro cant do w and h scaling while rotating, so we average the two
    if(rotation)
    {
        float scale = (wchange + hchange) / 2;
    
        rotate_scaled_sprite(softwareRenderer.screen, bitmap, x, y, ftofix(rotation * 0xff), ftofix(scale));
    }
    else
    {
        masked_stretch_blit(bitmap, softwareRenderer.screen, 0, 0, w, h, x, y, int(w * wchange), int(h * hchange));
    }
}
