#include <alleggl.h>
#ifdef WIN32
#include <winalleg.h>
#endif
#include "conf.h"
#include "log.h"
#include "datafile.h"
#include "settings.h"
#include "print.h"
#include "timer.h"
#include "sounds.h"
#include "animation.h"
#include "software.h"
#include "stopwatch.h"
#include "thread.h"
#include <allegro.h>
#include <iostream>
#include <sstream>
using namespace std;                    

#ifndef ENGINE_NAME
#define ENGINE_NAME "Engine"
#endif

namespace {
    
    bool allegro_usable = false;
    volatile bool switchedIn = true;
    volatile bool switchOutAsync = false;
    volatile bool switchInAsync = false;
    volatile bool quitAsync = false;
}
    
static void onSwitchIn()
{
    switchedIn = true;
    switchInAsync = true;
}
END_OF_FUNCTION(onSwitchIn)

static void onSwitchOut()
{
    switchedIn = false;
    switchOutAsync = true;
}
END_OF_FUNCTION(onSwitchOut)

static void onQuitRequest()
{
    quitAsync = true;
}
END_OF_FUNCTION(onQuitRequest)

Conf *globalConfInstance = 0;
namespace {
    
    int loadEverything(void*)
    {
        
        out << "Loading datafile.\n";
        datafile.load();
        out << "Datafile loaded.\n";
        packfile_password(0);
        out << "Loading sound effects.\n";
        Sounds::load();
        out << "Sound effects loaded.\n";
        
        return 0;
    }
}

const int Conf::WINDOWED = 1;
const int Conf::FULLSCREEN = 2;

const int Conf::SOFTWARE = 1;
const int Conf::HARDWARE = 2;

Conf::Conf(Flags flags, Mode renderingMode_) throw (const char*)
    : renderingMode(renderingMode_)
{
    if(globalConfInstance)
        throw "Multiple Conf instances are not allowed.";
    
    globalConfInstance = this;
    
#ifndef DEBUGMODE
    three_finger_flag = FALSE; // Disable control+alt+end ending the program for release versions
#endif
    
    LOCK_FUNCTION(onSwitchIn);
    LOCK_FUNCTION(onSwitchOut);
    LOCK_FUNCTION(onQuitRequest);
    
    set_window_title(ENGINE_NAME);
    set_close_button_callback(onQuitRequest);
    
    allegro_usable = true;
    
    if(0 != install_keyboard())
        throw "Unable to access keyboard.";
    
    if(install_mouse() < 0)
        throw "Unable to access mouse.";
    enable_hardware_cursor();
    show_os_cursor(MOUSE_CURSOR_ARROW);
    
    if(0 != install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, 0)) {
        
        string error = allegro_error;
        
        if(0 != install_sound(DIGI_AUTODETECT, MIDI_NONE, 0)) {
            
            error = allegro_error;
            
            if(0 != install_sound(DIGI_NONE, MIDI_AUTODETECT, 0))
                err << "Warning: Unable to install any sound drivers.  " << allegro_error << "\n";
            else
                err << "Warning: Unable to install DIGI sound drivers.  " << allegro_error << "\n";
        }
        else
            err << "Warning: Unable to install MIDI sound drivers.  " << error << "\n";
    }
    Bitmap::init();
    
    if(renderingMode == HARDWARE)
        initHardware(flags);
    else
        initSoftware(flags);
    
    animation cd("misc/cd/");
    animation loading("misc/loading/");
    
    Timer::time_t latestTime = timer.msecs();
    Tint tint(1, 1, 1, 1, 0);
    bool fadingIn = true;
    bool loadingFont = false;
    bool done = false;
    float diff;
    float rotation = 0;
    const unsigned int fadeTime = 800; // In milliseconds
    
    Thread loadThread(loadEverything);
    
    while(!done) {
        
        if(loadThread.done() && !loadingFont) {
            
            out << "Loading font.\n";
            Print::startInit();
            loadingFont = true;
        }
        
        if(loadingFont) {
            
            if(!Print::pollInit()) {
                
                out << "Font loaded.\n";
                done = true;
            }
        }
        
        if(timer.msecs() - latestTime > fadeTime) {
            
            fadingIn = !fadingIn;
            latestTime = timer.msecs();
        }
        
        diff = (timer.msecs() - latestTime) / (float)(fadeTime);
        
        rotation += diff / 2;
        
        if(fadingIn)
            tint.a = diff;
        else
            tint.a = 1 - diff;
        
        clearScreen();
        
        if(cd.get_bmp())
            cd.get_bmp()->blitToScreen(320 - cd.get_bmp()->getw() / 2, 240 - cd.get_bmp()->geth() + 20);
        
        if(loading.get_bmp())
            loading.get_bmp()->blitToScreen(320 - loading.get_bmp()->getw() / 2, 240 - loading.get_bmp()->geth() / 2, tint);
        
        flip();
    }
}

Conf::~Conf()
{
    Sounds::unload();
    Print::shutdown();
    datafile.unload();
    Bitmap::shutdown();
    
    if(renderingMode == HARDWARE)
        shutdownHardware();
    else
        shutdownSoftware();
    
    allegro_exit();
    
    allegro_usable = false;
    
    globalConfInstance = 0;
}

bool Conf::changeResolution(Flags flags, int w, int h, int depth)
{
    if(renderingMode == HARDWARE)
        return changeResolutionHardware(flags, w, h, depth);
    else
        return changeResolutionSoftware(flags, w, h, depth);
}

void Conf::flip()
{
    if(renderingMode == HARDWARE)
        flipHardware();
    else
        flipSoftware();
}

void Conf::clearScreen()
{
    if(renderingMode == HARDWARE)
        clearScreenHardware();
    else
        clearScreenSoftware();
}

bool Conf::switchedIn()
{
    return ::switchedIn;
}

bool Conf::switchedInAsync()
{
    int ret = switchInAsync;
    
    switchInAsync = false;
    
    return ret;
}

bool Conf::switchedOutAsync()
{
    int ret = switchOutAsync;
    
    switchOutAsync = false;
    
    return ret;
}

bool Conf::requestedQuitAsync()
{
    return quitAsync;
}

Conf::Mode Conf::getRenderingMode()
{
    return renderingMode;
}

void Conf::initSoftware(Flags)
{
    out << "Trying to load software renderer.\n";
    
    set_color_depth(settings.colorDepth);
    
    if(settings.fullscreen) {
        if(0 != softwareRenderer.init(GFX_AUTODETECT_FULLSCREEN, settings.screenWidth, settings.screenHeight, SoftwareRenderer::page_flip))
            throw "Unable to set software graphics mode.";
    }
    else if(0 != softwareRenderer.init(GFX_AUTODETECT_WINDOWED, settings.screenWidth, settings.screenHeight, SoftwareRenderer::page_flip))
        throw "Unable to set software graphics mode.";
    
    if(0 != set_display_switch_mode(SWITCH_BACKGROUND))
        if(0 != set_display_switch_mode(SWITCH_BACKAMNESIA))
            err << "Warning: Unable to set backgrounding switching mode.\n";
    
    if(0 != set_display_switch_callback(SWITCH_IN, onSwitchIn))
        err << "Warning: Unable to detect switching into program.\n";
    
    if(0 != set_display_switch_callback(SWITCH_OUT, onSwitchOut))
        err << "Warning: Unable to detect switching away from program.\n";
    
    out << "Software Renderer loaded successfully.\n" << softwareRenderer.get_mode() << ".\n";
}

void Conf::flipSoftware()
{
    softwareRenderer.post_draw();
}

void Conf::clearScreenSoftware()
{
    clear(softwareRenderer.screen);
}

bool Conf::changeResolutionSoftware(Flags, int, int, int)
{
    return true;
}

void Conf::shutdownSoftware()
{
    softwareRenderer.shutdown();
}

void Conf::initHardware(Flags)
{
    out << "Trying to load OpenGL renderer.\n";
    
    if(0 != install_allegro_gl())
        throw "Unable to init Allegro GL.";
    
    allegro_gl_clear_settings();
    allegro_gl_set(AGL_DOUBLEBUFFER, 1);
    allegro_gl_set(AGL_RENDERMETHOD, 1);
    allegro_gl_set(AGL_COLOR_DEPTH, settings.colorDepth);
	allegro_gl_set(AGL_Z_DEPTH, 16);
    allegro_gl_set(AGL_SAMPLE_BUFFERS, 0);
    allegro_gl_set(AGL_SAMPLES, 0);
    allegro_gl_set(AGL_SUGGEST, AGL_DOUBLEBUFFER | AGL_COLOR_DEPTH
        | AGL_RENDERMETHOD | AGL_SAMPLE_BUFFERS | AGL_SAMPLES);
    
    set_color_depth(settings.colorDepth);
    
    int mode;
    
    if(settings.fullscreen)
        mode = GFX_OPENGL_FULLSCREEN;
    else {
        mode = GFX_OPENGL_WINDOWED;
		allegro_gl_set(AGL_WINDOWED, TRUE);
		allegro_gl_set(AGL_REQUIRE, AGL_WINDOWED);
	}
    
    if(0 != set_gfx_mode(mode, settings.screenWidth, settings.screenHeight, 0, 0))
        throw "Unable to set OpenGL graphics mode.";
    
    if(0 != set_display_switch_mode(SWITCH_BACKGROUND))
        if(0 != set_display_switch_mode(SWITCH_BACKAMNESIA))
            err << "Warning: Unable to set backgrounding switching mode.\n";
    
    if(0 != set_display_switch_callback(SWITCH_IN, onSwitchIn))
        err << "Warning: Unable to detect switching into program.\n";
    
    if(0 != set_display_switch_callback(SWITCH_OUT, onSwitchOut))
        err << "Warning: Unable to detect switching away from program.\n";
    
    allegro_gl_begin();
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(0.0, 640.0, 480.0, 0.0, -1.0, 1.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glClearColor(0, 0, 0, 1);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glDisable(GL_DEPTH_TEST);
        allegro_gl_use_alpha_channel(1);
        allegro_gl_set_texture_format(GL_RGBA8);
        glEnable(GL_BLEND);
        glEnable(GL_TEXTURE_2D);
    allegro_gl_end();
    
    out << "OpenGL Renderer loaded successfully.\n";
}

void Conf::flipHardware()
{
    allegro_gl_flip();
}

void Conf::clearScreenHardware()
{
    glFlush();
    glClear(GL_COLOR_BUFFER_BIT);
}

bool Conf::changeResolutionHardware(Flags flags, int w, int h, int depth)
{
    out << "Changing res to " << flags << ", " << w << "x" << h << "@" << depth << "\n";
    
    Bitmap::unloadAll();
    
    allegro_gl_clear_settings();
    allegro_gl_set(AGL_DOUBLEBUFFER, 1);
    allegro_gl_set(AGL_RENDERMETHOD, 1);
    allegro_gl_set(AGL_COLOR_DEPTH, settings.colorDepth);
	allegro_gl_set(AGL_Z_DEPTH, 16);
    allegro_gl_set(AGL_SAMPLE_BUFFERS, 0);
    allegro_gl_set(AGL_SAMPLES, 0);
    allegro_gl_set(AGL_SUGGEST, AGL_DOUBLEBUFFER | AGL_COLOR_DEPTH
        | AGL_RENDERMETHOD | AGL_SAMPLE_BUFFERS | AGL_SAMPLES);
    
    set_color_depth(settings.colorDepth);
    
    int mode;
    
    if(settings.fullscreen)
        mode = GFX_OPENGL_FULLSCREEN;
    else {
        mode = GFX_OPENGL_WINDOWED;
		allegro_gl_set(AGL_WINDOWED, TRUE);
		allegro_gl_set(AGL_REQUIRE, AGL_WINDOWED);
	}
    
    if(0 != set_gfx_mode(mode, settings.screenWidth, settings.screenHeight, 0, 0))
        return false;
    
    allegro_gl_begin();
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(0.0, 640.0, 480.0, 0.0, -1.0, 1.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glClearColor(0, 0, 0, 1);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glDisable(GL_DEPTH_TEST);
        allegro_gl_use_alpha_channel(1);
        allegro_gl_set_texture_format(GL_RGBA8);
        glEnable(GL_BLEND);
        glEnable(GL_TEXTURE_2D);
    allegro_gl_end();
    
    out << "Res change successfull.\n";
    
    Bitmap::reloadAll();
    
    return true;
}

void Conf::shutdownHardware()
{
    remove_allegro_gl();
}

void Conf::handleCaughtError(const string &error)
{
    ostringstream ss;
    
    ss << error << endl << allegro_error << endl;
    
    string message = ss.str();
    
    if(allegro_usable) {
        
        set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
        allegro_message("%s", message.c_str());
    }
    
    err << message;
}
